/* 0S01-HEIDENHAIN_TE-355-A_Panel-002
  Program slouzi ke cteni matice tlacitek 8x8 a dvou potenciometru
*/

#include "EasyCAT.h"                // EasyCAT library to interface the LAN9252
#include <SPI.h>                    // SPI library


EasyCAT EASYCAT;                    // EasyCAT instance

// poznamky ponechany ze vzoroveho prikladu
// The constructor allow us to choose the pin used for the EasyCAT SPI chip select
// Without any parameter pin 9 will be used

// for EasyCAT board REV_A we can choose between:
// 8, 9, 10
//
// for EasyCAT board REV_B we can choose between:
// 8, 9, 10, A5, 6, 7

// example:
//EasyCAT EASYCAT(8);               // pin 8 will be used as SPI chip select


// The chip select chosen by the firmware must match the setting on the board

// On board REV_A the chip select is set soldering
// a 0 ohm resistor in the appropriate position

// On board REV_B the chip select is set
// througt a bank of jumpers




//definovani pinu
const int ledPin = 13;  // na tomto pinu je led-ka pro kontrolni blikani

// definovani casovych promenych
unsigned long cas;
unsigned long casKontrolLED;
unsigned long casVypisu;


// stav kontrolní ledky
boolean stavKontrolLED;


// promene vyhodnocovane matice tlacitek
byte maticeRadek = 0;
unsigned int maticeRadekPin[8];
byte maticeSloupec = 0;
unsigned int maticeSloupecPin[8];
byte maticeTlacitek[8];  // promena do ktere se zapisuje stav jednotlivych klaves

// promene potenciometru
const int spindlePin = A14;                // potenciometr S
byte spindle;
const int feedPin = A15;                   // potenciometr F
byte feed;

void setup()   // tento blok se provede pouze jednou
{
  Serial.begin(9600);  // spusteni seriove komunikace

  //definovani pinu maticove klavesnice
  maticeRadekPin[0] = 22;
  maticeRadekPin[1] = 23;
  maticeRadekPin[2] = 24;
  maticeRadekPin[3] = 25;
  maticeRadekPin[4] = 26;
  maticeRadekPin[5] = 27;
  maticeRadekPin[6] = 28;
  maticeRadekPin[7] = 29;

  maticeSloupecPin[0] = 30;
  maticeSloupecPin[1] = 31;
  maticeSloupecPin[2] = 32;
  maticeSloupecPin[3] = 33;
  maticeSloupecPin[4] = 34;
  maticeSloupecPin[5] = 35;
  maticeSloupecPin[6] = 36;
  maticeSloupecPin[7] = 37;


  // nastaveni pinu
  pinMode(ledPin, OUTPUT);

  for (int i = 0; i <= 7; i++)
  {
    pinMode(maticeRadekPin[i], INPUT);
    pinMode(maticeSloupecPin[i], INPUT);

  }

  // Popis programu odeslany po seriove lince pri spusteni
  Serial.println("Nacten program: 0S01-HEIDENHAIN_TE-355-A_Panel-002.ino");
  Serial.println("Program slouzi ke cteni matice tlacitek 8x8");
  Serial.println("a dvou potenciometru feed a spindle");
  Serial.println(); //Vypsani prazdneho radku

  Serial.print("Pin  1 na konektoru panelu je zapojen na arduinu do pinu ");
  Serial.println(maticeRadekPin[0]);
  Serial.print("Pin  2 na konektoru panelu je zapojen na arduinu do pinu ");
  Serial.println(maticeRadekPin[1]);
  Serial.print("Pin  3 na konektoru panelu je zapojen na arduinu do pinu ");
  Serial.println(maticeRadekPin[2]);
  Serial.print("Pin  4 na konektoru panelu je zapojen na arduinu do pinu ");
  Serial.println(maticeRadekPin[3]);
  Serial.print("Pin  5 na konektoru panelu je zapojen na arduinu do pinu ");
  Serial.println(maticeRadekPin[4]);
  Serial.print("Pin  6 na konektoru panelu je zapojen na arduinu do pinu ");
  Serial.println(maticeRadekPin[5]);
  Serial.print("Pin  7 na konektoru panelu je zapojen na arduinu do pinu ");
  Serial.println(maticeRadekPin[6]);
  Serial.print("Pin  8 na konektoru panelu je zapojen na arduinu do pinu ");
  Serial.println(maticeRadekPin[7]);
  Serial.println(); //Vypsani prazdneho radku

  Serial.print("Pin 20 na konektoru panelu je zapojen na arduinu do pinu ");
  Serial.println(maticeSloupecPin[0]);
  Serial.print("Pin 21 na konektoru panelu je zapojen na arduinu do pinu ");
  Serial.println(maticeSloupecPin[1]);
  Serial.print("Pin 22 na konektoru panelu je zapojen na arduinu do pinu ");
  Serial.println(maticeSloupecPin[2]);
  Serial.print("Pin 23 na konektoru panelu je zapojen na arduinu do pinu ");
  Serial.println(maticeSloupecPin[3]);
  Serial.print("Pin 24 na konektoru panelu je zapojen na arduinu do pinu ");
  Serial.println(maticeSloupecPin[4]);
  Serial.print("Pin 25 na konektoru panelu je zapojen na arduinu do pinu ");
  Serial.println(maticeSloupecPin[5]);
  Serial.print("Pin 26 na konektoru panelu je zapojen na arduinu do pinu ");
  Serial.println(maticeSloupecPin[6]);
  Serial.print("Pin 27 na konektoru panelu je zapojen na arduinu do pinu ");
  Serial.println(maticeSloupecPin[7]);
  Serial.println(); //Vypsani prazdneho radku


  //---- inicializace EasyCAT boardu -----

  // pokud je initializace uspesne kompletni
  if (EASYCAT.Init() == true)
  {
    Serial.print ("initialized");
  }

  else
    // initializace se nezdarila
  {
    Serial.print ("initialization failed");
    // The most common reason is that the SPI
    // chip select choosen on the board doesn't
    // match the one choosen by the firmware

    pinMode(13, OUTPUT);
    // stay in loop for ever
    // with the Arduino led blinking
    while (1)
    { //
      digitalWrite (13, LOW);
      delay(100);
      digitalWrite (13, HIGH);
      delay(100);
    }
  }
}

void loop()  // tento blok se bude provadet stale dokola
{

  cas = millis();  // nacteni casu vnitrniho casovace do promene cas

  aktivaceVstupuMaticeTlacitek ();
  // zde nesmi okamzite nasledovat vyhodnoceniMaticeTlacitek();

  delayMicroseconds(100);  // Pokud toto spozdeni tady nebylo, vyhodnocovalo to spatne tlacítka
  // bez problemu to fungovalo i při 10 us.

  vyhodnoceniPotenciometru ();

  vyhodnoceniMaticeTlacitek();

  // vypsaniStavuTlacitekPotenciometru (); // funkce pro ladeni, vypisuje pres seriovou linku stav tlacitek dvojkovem zapisu a hodnotu BYTE dvou potenciometru

  if (  maticeSloupec == 0 & maticeRadek == 0)// nahrada za cekani 2 milisekundy kvuli EasyCATu (pozadavek na cekani byl to v prikladu)
    // uloha EasyCATu se provede, až se zkontroluje stav všech tlačítek
  {
    EASYCAT.MainTask();   // provedeni ulohy EasyCAT
  }

  kontrolniLED(); // kontrolni blikani pro zobrazeni funkce Arduina, zkratka informace, ze se arduino nekouslo

}

void kontrolniLED()
{
  if (casKontrolLED < cas)
  {
    casKontrolLED = cas + 500;

    if (stavKontrolLED == true)
    {
      digitalWrite(ledPin, HIGH); // rozsviceni kontrolni LED
      stavKontrolLED = false;
    }
    else
    {
      digitalWrite(ledPin, LOW); // zhasnuti kontrolni LED
      stavKontrolLED = true;
    }
  }
}

void aktivaceVstupuMaticeTlacitek ()
{
  pinMode(maticeRadekPin[maticeRadek], OUTPUT);
  digitalWrite(maticeRadekPin[maticeRadek], LOW);
  pinMode(maticeSloupecPin[maticeSloupec], INPUT_PULLUP);
}

void vyhodnoceniMaticeTlacitek()
{

  if (digitalRead(maticeSloupecPin[maticeSloupec]) == LOW)   // pokud je zaregistrovani stlaceni tlacitka
  {
    // maticeTlacitek[maticeRadek]  |= (1 << maticeSloupec);        // dojde k zapsani bitu do BYTE z pole maticeTlacitek
    // nutne uvolnit pro vypsaniStavuTlacitekPotenciometru ()
    EASYCAT.BufferIn.Byte[maticeRadek] |= (1 << maticeSloupec);  // dojde k zapsani bitu do bufferu EasyCATu
  }
  else
  {
    // maticeTlacitek[maticeRadek]  &= ~(1 << maticeSloupec);
    // nutne uvolnit pro vypsaniStavuTlacitekPotenciometru ()
    EASYCAT.BufferIn.Byte[maticeRadek] &= ~(1 << maticeSloupec);
  }

  pinMode(maticeRadekPin[maticeRadek], INPUT);           //deaktivace rezimu sink na pinu pro radek
  pinMode(maticeSloupecPin[maticeSloupec], INPUT);       //deaktivace pull up rezistoru na pinu pro sloupec



  maticeRadek = maticeRadek + 1 ;          // posunuti se na dalsi radek

  if (maticeRadek == 8)                    // Vraceni se na nulty radek a posunuti sloupecku
  {
    maticeRadek = 0;

    maticeSloupec = maticeSloupec + 1;     // posunuti se na dalsi sloupec
    if (maticeSloupec == 8)
    {
      maticeSloupec = 0;
    }
  }
}

void vyhodnoceniPotenciometru ()
{
  //---------------------- Vzorovy zapis hodnot potenciometru do EasyCATu-----------
  // UWORD Analog0;
  // UWORD Analog1;

  // --- analog inputs management ---
  // Analog0.Word = analogRead(feedPin);                    // read analog input 0
  // Analog0.Word >>= 2;                                 // normalize it on 8 bits
  // EASYCAT.BufferIn.Byte[0] = Analog0.Byte[0];         // and put the result into

  // Analog1.Word = analogRead(spindlePin);                    // read analog input 1
  // Analog1.Word >>= 2;                                 // normalize it on 8 bits
  //EASYCAT.BufferIn.Byte[1] = Analog1.Byte[0];         // and put the result into
  //---------------------- Vzorovy zapis hodnot potenciometru do EasyCATu-----------


  feed = analogRead(feedPin) / 4; // funkce analogRead ma 10-ti bitovy vystup, proto je hodnota podelena 4
  EASYCAT.BufferIn.Byte[8] = feed;

  spindle = analogRead(spindlePin) / 4;
  EASYCAT.BufferIn.Byte[9] = spindle;


}
/*
  void vypsaniStavuTlacitekPotenciometru ()  // funkce pouze pro ladeni, v loop() by mela byt potlacena
  {
  if (casVypisu < cas)
  {
    casVypisu = cas + 100;
    for (int i = 0; i <= 7; i++)
    {
      for (int u = 0; u <= 7; u++)
      {
        boolean stav;
        if (maticeTlacitek[i] & (1 << u))
          stav = true;
        else
          stav = false;
        Serial.print(stav);
      }

      Serial.print(" ");
    }
    Serial.print(feed);
    Serial.print(" ");
    Serial.print(spindle);
    Serial.println();
  }
  }
*/
